(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright (c) 2013-2014 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: display_window.ml,v 1.19 2014/07/16 21:46:59 fclement Exp $ *)

(* The graphics window is split into sub-windows. Each sub-window is
 surrounded by a white margin. Let n be the dimension of the parameter.
 Each group of n sub-windows is decorated with a title and a colorbar.
 
 Parameter_graphics -> n x 1 or 1 x n sub-windows,
 Full_graphics -> n x 2 or 2 x n sub-windows.
 
 Parameter_graphics means that all components of the parameter are displayed.
 Hence, n sub-windows.
 
 Full_graphics means that all components for both the parameter and the
 gradient of the cost function are displayed.
 Hence, 2n sub-windows.
 
 Depending on the aspect ratio of the domain, component sub-windows are
 either displayed along lines or columns:
 for Parameter_graphics,
   Portrait -> n x 1 sub-windows
   Landscape -> 1 x n sub-windows,
 for Full_graphics,
   Portrait -> n x 2 sub-windows
   Landscape -> 2 x n sub-windows. *)

module Nat = Unatural;;
module Prational = Upositive_rational;;
module Prat_range = Upositive_rational_range;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
and _3_ = Nat._3_
and _4_ = Nat._4_
and _5_ = Nat._5_
;;

open Upositive_rational_range_types;;

module Config = Configuration;;

let dw_qual = Upervasives.qualify "Display_window";;

let dw_invalid_arg = Upervasives.invalid_arg_qual dw_qual;;
let dw_print_debug fname = Config.print_debug (dw_qual fname);;

open Refinement_types;;
open Display_types;;

let ppf = Config.get_formatter ();;

let print_nat = Unatural_print.print_t;;
let print_rat = Upositive_rational_range_types_print.print_t;;

(* Compound graphics window. *)

let add_margins m x =
  let two_ms = Nat.mul _2_ m in
  Nat.add x two_ms
;;

let same_direction dir comps_dir =
  match (dir, comps_dir) with
  | (Horizontal_dir, Horizontal_comps_dir)
  | (Vertical_dir, Vertical_comps_dir) -> true
  | (Horizontal_dir, Vertical_comps_dir)
  | (Vertical_dir, Horizontal_comps_dir) -> false
;;

let aspect (w, h) = if w <= h then Portrait else Landscape;;

let comps_direction sz =
  match aspect sz with
  | Portrait -> Horizontal_comps_dir
  | Landscape -> Vertical_comps_dir
;;

(* Height of the title sub-window. *)
let title_sub_window_height m (_, ft_h) = add_margins m ft_h;;

(* Polar size of the legend sub-window. *)
let legend_sub_window_polar_size comps_dir m (ft_w, ft_h) nbc =
  let (leg_w, leg_h) = (Nat.mul nbc ft_w, ft_h) in
  let (leg_t, leg_n) =
    match comps_dir with
    | Horizontal_comps_dir -> (leg_w, leg_h)
    | Vertical_comps_dir -> (leg_h, leg_w) in
  (add_margins m leg_t, add_margins m leg_n)
;;

(* Size of the colorbar. *)
let colorbar_size comps_dir (nx, ny) cb_t =
  match comps_dir with
  | Horizontal_comps_dir -> (cb_t, ny)
  | Vertical_comps_dir -> (nx, cb_t)
;;

(* Tangential length of the colorbar sub_window. *)
let colorbar_sub_window_tangential m cb_t = add_margins m cb_t;;

(* Tangential length of the colormap sub-window. *)
let colormap_sub_window_tangential comps_dir m cb_t ft_sz nbc =
  let cb_sw_t = colorbar_sub_window_tangential m cb_t in
  let (leg_sw_t, _) = legend_sub_window_polar_size comps_dir m ft_sz nbc in
  Nat.add cb_sw_t leg_sw_t
;;

(* To compute the scaling factor, lengths are represented in the form
 alpha x (l+2m) + beta where l is the length of one pixmap and m the margin.
 Hence, in the following functions, lengths are abstracted as (alpha, beta). *)

(* Variable part for each quantity (= alpha). *)
let quantity_variable_part n dir comps_dir =
  if same_direction dir comps_dir then n else _1_
;;

(* Fixed part for each quantity (= beta). *)
let quantity_fixed_part dir comps_dir m cb_t ft_sz nbc =
  let tit_sw_h = title_sub_window_height m ft_sz in
  let cm_sw_t = colormap_sub_window_tangential comps_dir m cb_t ft_sz nbc in
  match (dir, comps_dir) with
  | (Horizontal_dir, Horizontal_comps_dir) -> cm_sw_t
  | (Horizontal_dir, Vertical_comps_dir) -> _0_
  | (Vertical_dir, Horizontal_comps_dir) -> tit_sw_h
  | (Vertical_dir, Vertical_comps_dir) -> Nat.add cm_sw_t tit_sw_h
;;

(* Alpha of the full window. *)
let window_alpha n gm dir comps_dir =
  let alpha = quantity_variable_part n dir comps_dir in
  match gm with
  | No_graphics -> assert false
  | Parameter_graphics -> alpha
  | Full_graphics ->
      if same_direction dir comps_dir then alpha else Nat.mul _2_ alpha
;;

(* Beta of the full window. *)
let window_beta gm dir comps_dir m cb_t ft_sz nbc =
  let beta = quantity_fixed_part dir comps_dir m cb_t ft_sz nbc in
  match gm with
  | No_graphics -> assert false
  | Parameter_graphics -> beta
  | Full_graphics ->
      begin
        match (dir, comps_dir) with
        | (Horizontal_dir, Horizontal_comps_dir)
        | (Horizontal_dir, Vertical_comps_dir)
        | (Vertical_dir, Vertical_comps_dir) -> beta
        | (Vertical_dir, Horizontal_comps_dir) -> Nat.mul _2_ beta
      end
;;

(* Length of the pixmap sub-windows. *)
let pixmap_sub_window_length m l = add_margins m l;;

(* Length of the pixmaps sub-window. *)
let pixmaps_sub_window_length n dir comps_dir m l =
  let alpha = quantity_variable_part n dir comps_dir in
  let pxm_sw_l = pixmap_sub_window_length m l in
  Nat.mul alpha pxm_sw_l
;;

(* Length of the quantity sub-window. *)
let quantity_sub_window_length n dir comps_dir m l cb_t ft_sz nbc =
  let pxms_sw_l = pixmaps_sub_window_length n dir comps_dir m l in
  let beta = quantity_fixed_part dir comps_dir m cb_t ft_sz nbc in
  Nat.add pxms_sw_l beta
;;

(* nn = alpha*(l + 2*m) + beta. *)
let window_length n gm dir comps_dir m l cb_t ft_sz nbc =
  let alpha = window_alpha n gm dir comps_dir in
  let pxm_sw_l = pixmap_sub_window_length m l in
  let pxms_sw_l = Nat.mul alpha pxm_sw_l in
  let beta = window_beta gm dir comps_dir m cb_t ft_sz nbc in
  Nat.add pxms_sw_l beta
;;

(* Solve previous formula with l <- r*l for r.
 r = (nn - beta - alpha*2*m) / (alpha*l). *)
let split_ratio n gm dir comps_dir m l cb_t ft_sz nbc nn =
  let two_ms = Nat.mul _2_ m in
  let alpha = window_alpha n gm dir comps_dir in
  dw_print_debug "split_ratio" "alpha" print_nat ppf alpha;
  let beta = window_beta gm dir comps_dir m cb_t ft_sz nbc in
  dw_print_debug "split_ratio" "beta" print_nat ppf beta;
  let p = Nat.sub (Nat.sub nn beta) (Nat.mul alpha two_ms)
  and q = Nat.mul alpha l in
  Prational.of_nat p q
;;

(* Size of the pixmap sub-windows. *)
let pixmap_sub_window_size m (nx, ny) =
  (pixmap_sub_window_length m nx, pixmap_sub_window_length m ny)
;;

(* Size of the pixmaps sub-window. *)
let pixmaps_sub_window_size n comps_dir m (nx, ny) =
  (pixmaps_sub_window_length n Horizontal_dir comps_dir m nx,
   pixmaps_sub_window_length n Vertical_dir comps_dir m ny)
;;

(* Size of the quantity sub-window. *)
let quantity_sub_window_size n comps_dir m (nx, ny) cb_t ft_sz nbc =
  (quantity_sub_window_length n Horizontal_dir comps_dir m nx cb_t ft_sz nbc,
   quantity_sub_window_length n Vertical_dir comps_dir m ny cb_t ft_sz nbc)
;;

let margin = Nat.of_int 10;;

let colorbar_tangential = Nat.mul _3_ margin;;

let font = "10x20";;
let font_size = (Nat.of_int 10, Nat.of_int 20);;

(* Format for colormap bounds is "%-8.2g".*)
let leg_nb_char = Nat.of_int 8;;

let scaling_factor n gm (nnx_min, nny_min) (nnx_max, nny_max) (w, h) =
  let comps_dir = comps_direction (w, h) in
  let rgx =
    Prat_range.of_rat
      (split_ratio
         n gm Horizontal_dir comps_dir
         margin w colorbar_tangential font_size leg_nb_char nnx_min)
      (split_ratio
         n gm Horizontal_dir comps_dir
         margin w colorbar_tangential font_size leg_nb_char nnx_max)
  and rgy =
    Prat_range.of_rat
      (split_ratio
         n gm Vertical_dir comps_dir
         margin h colorbar_tangential font_size leg_nb_char nny_min)
      (split_ratio
         n gm Vertical_dir comps_dir
         margin h colorbar_tangential font_size leg_nb_char nny_max) in
  dw_print_debug "scaling_factor" "rgx" print_rat ppf rgx;
  dw_print_debug "scaling_factor" "rgy" print_rat ppf rgy;
  let rg = Prat_range.intersection rgx rgy in
  match (rgx, rgy, rg) with
  | (Empty, _, _) | (_, Empty, _) ->
      dw_invalid_arg "scaling_factor" "incompatible window size bounds"
  | (Nonempty rgx, Nonempty rgy, Empty) ->
      Prational.min rgx.right rgy.right
  | (Nonempty _, Nonempty _, Nonempty _) ->
      Prat_range.project rg Prational._1_
;;

(* Graphics_elements. *)

let init_graphics_elements =
  {
   comps_direction = Horizontal_comps_dir;
   scaling_factor = Prational._1_;
   display_image_size = (_0_, _0_);
   window_size = (_0_, _0_);
   sub_window_sizes =
     {
      title_sub_window_height = _0_;
      legend_sub_window_polar_size = (_0_, _0_);
      colorbar_size = (_0_, _0_);
      colorbar_sub_window_tangential = _0_;
      colormap_sub_window_tangential = _0_;
      pixmap_sub_window_size = (_0_, _0_);
      quantity_sub_window_size = (_0_, _0_);
     };
  }
;;

(* This should be an abstract reference with unique assignment, weakening it to
 mandatory assignment is necessary for testing purposes though... *)
let { Uaref.get = get_graphics_elements; set = gen_set_graphics_elements; } =
  Uaref.make_aref_mandatory "graphics elements" init_graphics_elements
;;

let set_graphics_elements n gm min_w_sz max_w_sz ((w, h) as sz) =
  let comps_dir = comps_direction (w, h) in
  let sf = scaling_factor n gm min_w_sz max_w_sz sz in
  let scale = Prational.scale sf in
  let ((nx, ny) as d_sz) = (scale w, scale h) in
  let nnx =
    window_length
      n gm Horizontal_dir comps_dir
      margin nx colorbar_tangential font_size leg_nb_char
  and nny =
    window_length
      n gm Vertical_dir comps_dir
      margin ny colorbar_tangential font_size leg_nb_char in
  let tit_sw_h = title_sub_window_height margin font_size in
  let leg_sw_psz =
    legend_sub_window_polar_size comps_dir margin font_size leg_nb_char in
  let cb_sz = colorbar_size comps_dir d_sz colorbar_tangential in
  let cb_sw_t = colorbar_sub_window_tangential margin colorbar_tangential in
  let cm_sw_t =
    colormap_sub_window_tangential
      comps_dir margin colorbar_tangential font_size leg_nb_char in
  let pxm_sw_sz = pixmap_sub_window_size margin d_sz in
  let q_sw_sz =
    quantity_sub_window_size
      n comps_dir margin d_sz colorbar_tangential font_size leg_nb_char in
  let sw_szs =
    {
     title_sub_window_height = tit_sw_h;
     legend_sub_window_polar_size = leg_sw_psz;
     colorbar_size = cb_sz;
     colorbar_sub_window_tangential = cb_sw_t;
     colormap_sub_window_tangential = cm_sw_t;
     pixmap_sub_window_size = pxm_sw_sz;
     quantity_sub_window_size = q_sw_sz;
    } in
  let ge =
    {
     comps_direction = comps_dir;
     scaling_factor = sf;
     display_image_size = d_sz;
     window_size = (nnx, nny);
     sub_window_sizes = sw_szs;
    } in
  gen_set_graphics_elements ge
;;

(* Display window setup. *)

let open_display (nnx, nny) =
  Graphics.open_graph "";
  Graphics.set_window_title
    (Printf.sprintf "%s monitoring"
       Ref_indic_version.pph_id_card.Upph_id_card.project_name);
  Graphics.resize_window (Nat.to_int nnx) (Nat.to_int nny);
  Graphics.set_font font;
;;

let window_setup n gm min_w_sz max_w_sz sz =
  set_graphics_elements n gm min_w_sz max_w_sz sz;
  let { window_size = w_sz; _; } = get_graphics_elements () in
  open_display w_sz;
;;
